/** 
 * Multithreaded local proxy for a server past a firewall. 
 * Usage: LocalProxy proxy-host proxy-port (local-port remote-host remote-port)+ 
 * 
 * Usage examples:
 * java LocalProxy www-proxy 80 4444 domain.com 143 
 *    local port 4444 --> domain.com:143 via www-proxy proxy server
 *
 * java LocalProxy www-proxy 1080 4444 domain.com 143 4445 domain.com 993
 *    local port 4444 --> domain.com:143  
 *    local port 4445 --> domain.com:993
 *
 * I use this program to connect to mail server past a company
 * proxy server. I set up a .bat file to run something like 
 * the first example above and then point my mail client (netscape) at 
 * IMAP server at localhost and port 4444.
 *
 * Caveat 1: I know this works for my company's proxy server, I don't 
 * know the exact other proxy flavors that will work. 
 *
 * Caveat 2: There is a lot of noise that gets printed to stdout. 
 * Comment out the prints if they bothers you, but I like it.
 *
 * Note: The way that the proxy works is that you make a connection to 
 * the proxy server, ignore all text (the until you see two consecutive 
 * new lines (\r\f or acs(13)asc(10)) then send the text:
 * 
 * CONNECT remotehost:remoteport HTTP/1.1\n\n
 *
 * Note 2: This program can also be used to monitor the traffic that 
 * between the client program and the remote-host. It is pretty frightening
 * how much gets sent in plaintext -- like mail passwords. To see the traffic
 * just uncomment the printBuffer command below. Note
 * that if you are transferring binary data the output is a mess.
 **/
import java.net.*;
import java.io.*;

public class LocalProxy extends Thread {

    // Static fields that hold the proxy information
    public static String s_proxyHost = null;
    public static int    s_proxyPort = -1;


    public static int proxyCount = 0;
    public static void main(String[] args) {
	if ((args.length < 5) ||
	    (((args.length - 2) % 3) != 0)) {
	    System.out.println("Usage: LocalProxy proxy-host proxy-port (localport remotehost remoteport)*");
	    System.exit(1);
	}
	
	s_proxyHost = args[0];
	s_proxyPort = Integer.parseInt(args[1]);
       	
	int numProxies = args.length / 3;
	for (int i=0; i<numProxies; i++) {
	    int localPort     = Integer.parseInt(args[2 + 3*i+0]);
	    String remoteHost = args[2+ 3*i+1];
	    int remotePort    = Integer.parseInt(args[2+ 3*i+2]);
	    System.out.println("MAIN: spawning proxy: localhost (" + localPort + ") --> " +
			       remoteHost + " (" + remotePort + ")");
	    new LocalProxy(localPort, remoteHost, remotePort).start();
	}
	System.out.println("MAIN: done.");
    }
    
    // Instance specific stuff
    int    thisProxyId;
    int    locPort;
    String remHost;
    int    remPort;
    // Sets up the proxy connection
    public LocalProxy(int lp, String rh, int rp) {
	this.locPort = lp;
	this.remHost = rh;
	this.remPort = rp;
	this.thisProxyId = proxyCount++;
	println("created");
    }

    /** 
     * Main entry point of the proxy -- the proxy is a thread.
     * sit in a forever loop waiting for connections...
     **/
    public void run() {
	ServerSocket serverSocket = null;
	try {
	    serverSocket = new ServerSocket(this.locPort);
	    println("creating a server on port " + this.locPort);
	} catch (IOException e) {
	    println("Could not listen on port: " + this.locPort);
	    return;
	} 
	
	// sit in this loop accepting connections...
	while (true) {
	    try {
		println("listening.");
		Socket s = serverSocket.accept();
		println("connection accepted, dispatching.");
		dispatchConnection(s);
	    } catch (IOException e) {
		println("Accept failed on port " + this.locPort);
		e.printStackTrace();
	    }
	}
    }
    /**
     * create an connection from the incoming socket to the 
     * remote host going through the proxy server and
     * spawn streamer threads to handle passing data.
     **/
    private void dispatchConnection(Socket incomingSocket) throws IOException {
	println("connecting...");
	Socket outgoingSocket = new Socket(s_proxyHost, s_proxyPort);
	// begin the connection to the dest server
	PrintWriter pw = new PrintWriter(outgoingSocket.getOutputStream(),true);
	
	// string used to open a connection with the destination host
	// via the HTTP Proxy
	String connectString = 
	    "CONNECT " + this.remHost + ":" + this.remPort + " HTTP/1.1" + "\n\n";
	pw.print(connectString);
	pw.flush();
	
	// clean proxy input stream until two new lines hit
	clearProxyInput(outgoingSocket);
	
	println("starting streamers...");
	new ProxyStreamer(incomingSocket,
			  outgoingSocket).start();
	new ProxyStreamer(outgoingSocket,
			  incomingSocket).start();
	
	println("done spawning streamers.");
    }
    
    private void clearProxyInput(Socket s) 
	throws IOException {

	// now, read in data from the incoming socket, one byte at
	// at time until we get two new lines (13 10 13 10)
	byte[] buff = new byte[1];
	byte[] lastSeen = new byte[4];

	byte[] lookingFor = new byte[]{13, 10, 13, 10};

	InputStream osIn = s.getInputStream();

	int bytesRead = osIn.read(buff);
	boolean newLineFlag;
	do {
	    if (osIn.read(buff) != 1)
		throw new RuntimeException("error reading proxy header");
	    //System.out.println("proxy header: " + buff[0] + ":" + (char)buff[0]);
	    lastSeen = addToBuffer(lastSeen, buff[0]);
	} while (!compareBuffers(lastSeen, lookingFor));
	
    }

    // inefficient way to compare two byte arrays. Assume lengths ==
    private boolean compareBuffers(byte[] buf1, byte[] buf2) {
	for (int i=0; i<buf1.length; i++)
	    if (buf1[i] != buf2[i]) 
		return false;
	
	return true;
    }

    // adds a byte on to the end of the passed array, 
    // shifting other values forward
    private byte[] addToBuffer(byte[] inbuff, byte b) {
	byte[] outbuff = new byte[inbuff.length];
	for (int i=0; i<inbuff.length-1; i++)
	    outbuff[i] = inbuff[i+1];
	outbuff[inbuff.length-1] = b;
	return outbuff;
    }

    private void println(String s) {
	System.out.println("LocalProxy " + this.thisProxyId + ": " + s);
    }
}

// This thread reads from the input stream and writes to the output stream.
class ProxyStreamer extends Thread {
    static final int BUFFSIZE = 8*1024;
    static int streamerCounter = 0;

    Socket inSock;
    Socket outSock;
    int streamerId;
        
    public ProxyStreamer(Socket i, Socket o) {
	inSock = i;
	outSock = o;
	streamerId = streamerCounter++;
    }
    
    public void run() {
	InputStream is;
	OutputStream os;


	byte[] buff = new byte[BUFFSIZE];
	int bytesRead;
	try {
	    is = this.inSock.getInputStream();
	    os = this.outSock.getOutputStream();

	    while ((bytesRead = is.read(buff)) != -1) {
		os.write(buff,0,bytesRead);
		println("transferred " + bytesRead + " bytes");
		//printBuffer(buff, bytesRead);
	    }
	    
	    
	} catch (IOException e) {
	    println("streamer done:" + e.getMessage() + ".");
	} finally {
	    shutdown();
	}
    }
    
    // cleanup all system resources..
    void shutdown() {
	try {
	    this.inSock.close();
	} catch (IOException e) {
	    println("error closing insock");
	    e.printStackTrace();
	}

	try {
	    this.outSock.close();
	} catch (IOException e) {
	    println("error closing outsock");
	    e.printStackTrace();
	}
    }
    
    void println(String s) {
	System.out.println("streamer " + streamerId + ": " + s);
    }

    void printBuffer(byte buff[], int size) {
	String buffRep = "";
	for (int i=0; i<size; i++) {
	    buffRep += (char)buff[i];
	}
	println("buffer: " + buffRep);
    }
}
